@@ -13,7 +13,7 @@ use MIME::Base64 'encode_base64url';
 use Crypt::LE ':errors', ':keys';
 use utf8;
 
-my $VERSION = '0.33';
+my $VERSION = '0.38';
 
 exit main();
 
@@ -32,7 +32,17 @@ sub work {
     my $opt = shift;
     my $rv = parse_options($opt);
     return $rv if $rv;
-    my $le = Crypt::LE->new(autodir => 0, server => $opt->{'server'}, live => $opt->{'live'}, version => $opt->{'api'}||0, debug => $opt->{'debug'}, logger => $opt->{'logger'});
+    # Set the default protocol version to 2 unless it is set explicitly or custom server/directory is set (in which case auto-sense is used).
+    $opt->{'api'} = 2 unless (defined $opt->{'api'} or $opt->{'server'} or $opt->{'directory'});
+    my $le = Crypt::LE->new(
+	autodir => 0,
+	dir => $opt->{'directory'},
+	server => $opt->{'server'},
+	live => $opt->{'live'},
+	version => $opt->{'api'}||0,
+	debug => $opt->{'debug'},
+	logger => $opt->{'logger'},
+    );
 
     if (-r $opt->{'key'}) {
         $opt->{'logger'}->info("Loading an account key from $opt->{'key'}");
@@ -99,7 +109,6 @@ sub work {
             return $opt->{'error'}->("Could not load existing CSR key from $opt->{'csr-key'} - " . $le->error_details, 'CSR_KEY_LOAD') if $le->load_csr_key($opt->{'csr-key'});
         }
     } else {
-        return $opt->{'error'}->("For multi-webroot path usage, the amount of paths given should match the amount of domain names listed.", 'WEBROOT_MISMATCH') if _path_mismatch($le, $opt);
         $opt->{'logger'}->info("Generating a new CSR for domains $opt->{'domains'}");
         if (-e $opt->{'csr-key'}) {
              # Allow using pre-existing key when generating CSR
@@ -116,6 +125,7 @@ sub work {
             $opt->{'logger'}->info("Saving a new CSR key into $opt->{'csr-key'}");
             return $opt->{'error'}->("Failed to save a CSR key", 'CSR_SAVE') if _write($opt->{'csr-key'}, $le->csr_key);
         }
+        return $opt->{'error'}->("For multi-webroot path usage, the amount of paths given should match the amount of domain names listed.", 'WEBROOT_MISMATCH') if _path_mismatch($le, $opt);
     }
 
     return if $opt->{'generate-only'};
@@ -164,8 +174,8 @@ sub work {
     # Build a copy of the parameters from the command line and added during the runtime, reduced to plain vars and hashrefs.
     my %callback_data = map { $_ => $opt->{$_} } grep { ! ref $opt->{$_} or ref $opt->{$_} eq 'HASH' } keys %{$opt};
 
-    # We might not need to re-verify, verification holds for a while.
-    my $new_crt_status = $le->request_certificate();
+    # We might not need to re-verify, verification holds for a while. NB: Only do that for the standard LE servers.
+    my $new_crt_status = ($opt->{'server'} or $opt->{'directory'}) ? AUTH_ERROR : $le->request_certificate();
     unless ($new_crt_status) {
         $opt->{'logger'}->info("Received domain certificate, no validation required at this time.");
     } else {
@@ -196,32 +206,51 @@ sub work {
         $opt->{'logger'}->info("Requesting domain certificate.");
         return $opt->{'error'}->($le->error_details, 'CERTIFICATE_REQUEST') if $le->request_certificate();
     }
-    $opt->{'logger'}->info("Requesting issuer's certificate.");
-    if ($le->request_issuer_certificate()) {
-        $opt->{'logger'}->error("Could not download an issuer's certificate, " . ($le->issuer_url ? "try to download manually from " . $le->issuer_url : "the URL has not been provided."));
-        $opt->{'logger'}->warn("Will be saving the domain certificate alone, not the full chain.");
-        return $opt->{'error'}->("Failed to save the domain certificate file", 'CERTIFICATE_SAVE') if _write($opt->{'crt'}, $le->certificate);
+
+    my ($certificate, $issuer, $saved);
+
+    if ($opt->{'alternative'}) {
+        $opt->{'logger'}->info("Requesting alternative certificates.");
+        return $opt->{'logger'}->error($le->error_details, 'CERTIFICATE_REQUEST') if $le->request_alternatives();
+        if (my $alternative = $le->alternative_certificate($opt->{'alternative'} - 1)) {
+            ($certificate, $issuer) = @{$alternative};
+        } else {
+            return $opt->{'error'}->("There is no alternative certificate #$opt->{'alternative'}.", 'CERTIFICATE_REQUEST');
+        }
     } else {
+        $opt->{'logger'}->info("Requesting issuer's certificate.");
+        $certificate = $le->certificate;
+        if ($le->request_issuer_certificate()) {
+            $opt->{'logger'}->error("Could not download an issuer's certificate, " . ($le->issuer_url ? "try to download manually from " . $le->issuer_url : "the URL has not been provided."));
+            $opt->{'logger'}->warn("Will be saving the domain certificate alone, not the full chain.");
+            return $opt->{'error'}->("Failed to save the domain certificate file", 'CERTIFICATE_SAVE') if _write($opt->{'crt'}, $certificate);
+            $saved = 1;
+        } else {
+            $issuer = $le->issuer;
+        }
+    }
+
+    unless ($saved) {
         unless ($opt->{'legacy'}) {
             $opt->{'logger'}->info("Saving the full certificate chain to $opt->{'crt'}.");
-            return $opt->{'error'}->("Failed to save the domain certificate file", 'CERTIFICATE_SAVE') if _write($opt->{'crt'}, $le->certificate . "\n" . $le->issuer . "\n");
+            return $opt->{'error'}->("Failed to save the domain certificate file", 'CERTIFICATE_SAVE') if _write($opt->{'crt'}, $certificate . "\n" . $issuer . "\n");
         } else {
             $opt->{'logger'}->info("Saving the domain certificate to $opt->{'crt'}.");
-            return $opt->{'error'}->("Failed to save the domain certificate file", 'CERTIFICATE_SAVE') if _write($opt->{'crt'}, $le->certificate);
+            return $opt->{'error'}->("Failed to save the domain certificate file", 'CERTIFICATE_SAVE') if _write($opt->{'crt'}, $certificate);
             $opt->{'crt'}=~s/\.[^\.]+$//;
             $opt->{'crt'}.='.ca';
             $opt->{'logger'}->info("Saving the issuer's certificate to $opt->{'crt'}.");
-            $opt->{'logger'}->error("Failed to save the issuer's certificate, try to download manually from " . $le->issuer_url) if _write($opt->{'crt'}, $le->issuer);
+            $opt->{'logger'}->error("Failed to save the issuer's certificate", 'CERTIFICATE_SAVE') if _write($opt->{'crt'}, $issuer);
         }
     }
     if ($opt->{'export-pfx'}) {
         # Note: At this point the certificate is already issued, but with pfx export option active we will return an error if export has failed, to avoid triggering
         # the 'success' batch processing IIS users might have set up on issuance and export.
-        if ($le->issuer) {
+        if ($issuer) {
             my $target_pfx = $opt->{'crt'};
             $target_pfx=~s/\.[^\.]*$//;
             $opt->{'logger'}->info("Exporting certificate to $target_pfx.pfx.");
-            return $opt->{'error'}->("Error exporting pfx: " . $le->error_details, 'CERTIFICATE_EXPORT') if $le->export_pfx("$target_pfx.pfx", $opt->{'export-pfx'}, $le->certificate, $le->csr_key, $le->issuer, $opt->{'tag-pfx'});
+            return $opt->{'error'}->("Error exporting pfx: " . $le->error_details, 'CERTIFICATE_EXPORT') if $le->export_pfx("$target_pfx.pfx", $opt->{'export-pfx'}, $certificate, $le->csr_key, $issuer, $opt->{'tag-pfx'});
         } else {
             return $opt->{'error'}->("Issuer's certificate is not available, skipping pfx export to avoid creating an invalid pfx.", 'CERTIFICATE_EXPORT_ISSUER');
         }
@@ -229,7 +258,7 @@ sub work {
     if ($opt->{'complete-handler'}) {
         my $data = {
             # Note, certificate here is just a domain certificate, issuer is passed separately - so handler could merge those or use them separately as well.
-            certificate => $le->certificate, certificate_file => $opt->{'crt'}, key_file => $opt->{'csr-key'}, issuer => $le->issuer, 
+            certificate => $le->certificate, certificate_file => $opt->{'crt'}, key_file => $opt->{'csr-key'}, issuer => $le->issuer, alternatives => $le->alternative_certificates(),
             domains => $le->domains, logger => $opt->{'logger'},
         };
         my $rv;
@@ -241,7 +270,7 @@ sub work {
         }
     }
     $opt->{'logger'}->info("===> NOTE: You have been using the test server for this certificate. To issue a valid trusted certificate add --live option.") unless $opt->{'live'};
-    $opt->{'logger'}->info("The job is done, enjoy your certificate! For feedback and bug reports contact us at [ https://ZeroSSL.com | https://Do-Know.com ]\n");
+    $opt->{'logger'}->info("The job is done, enjoy your certificate!\n");
     return { code => $opt->{'issue-code'}||0 };
 }
 
@@ -249,9 +278,9 @@ sub parse_options {
     my $opt = shift;
     my $args = @ARGV;
 
-    GetOptions ($opt, 'key=s', 'csr=s', 'csr-key=s', 'domains=s', 'path=s', 'crt=s', 'email=s', 'curve=s', 'server=s', 'api=i', 'config=s', 'renew=i', 'renew-check=s','issue-code=i',
+    GetOptions ($opt, 'key=s', 'csr=s', 'csr-key=s', 'domains=s', 'path=s', 'crt=s', 'email=s', 'curve=s', 'server=s', 'directory=s', 'api=i', 'config=s', 'renew=i', 'renew-check=s','issue-code=i',
         'handle-with=s', 'handle-as=s', 'handle-params=s', 'complete-with=s', 'complete-params=s', 'log-config=s', 'update-contacts=s', 'export-pfx=s', 'tag-pfx=s',
-        'generate-missing', 'generate-only', 'revoke', 'legacy', 'unlink', 'delayed', 'live', 'quiet', 'debug+', 'help') ||
+        'alternative=i', 'generate-missing', 'generate-only', 'revoke', 'legacy', 'unlink', 'delayed', 'live', 'quiet', 'debug+', 'help') ||
         return $opt->{'error'}->("Use --help to see the usage examples.", 'PARAMETERS_PARSE');
 
     if ($opt->{'config'}) {
@@ -264,15 +293,20 @@ sub parse_options {
     my $rv = reconfigure_log($opt);
     return $rv if $rv;
 
-    $opt->{'logger'}->info("[ ZeroSSL Crypt::LE client v$VERSION started. ]");
+    $opt->{'logger'}->info("[ Crypt::LE client v$VERSION started. ]");
+    my $custom_server;
 
-    if ($opt->{'server'}) {
-        return $opt->{'error'}->("Unsupported protocol for the custom server URL: $1.", 'CUSTOM_SERVER_URL') if ($opt->{'server'}=~s~^(.*?)://~~ and uc($1) ne 'HTTPS');
-        my $server = $opt->{'server'}; # For logging.
-        $opt->{'logger'}->warn("Remember to URL-escape special characters if you are using server URL with basic auth credentials.") if $server=~s~[^@/]*@~~;
-        $opt->{'logger'}->info("Custom server URL 'https://$server' is used.");
-        $opt->{'logger'}->warn("Note: 'live' option is ignored.") if $opt->{'live'};
+    foreach my $url_type (qw<server directory>) {
+        if ($opt->{$url_type}) {
+            return $opt->{'error'}->("Unsupported protocol for the custom $url_type URL: $1.", 'CUSTOM_' . uc($url_type) . '_URL') if ($opt->{$url_type}=~s~^(.*?)://~~ and uc($1) ne 'HTTPS');
+            my $server = $opt->{$url_type}; # For logging.
+            $opt->{'logger'}->warn("Remember to URL-escape special characters if you are using $url_type URL with basic auth credentials.") if $server=~s~[^@/]*@~~;
+            $opt->{'logger'}->info("Custom $url_type URL 'https://$server' is used.");
+            $opt->{'logger'}->warn("Note: '$url_type' setting takes over the 'server' one.") if $custom_server;
+            $custom_server = 1;
+        }
     }
+    $opt->{'logger'}->warn("Note: 'live' option is ignored.") if ($opt->{'live'} and $custom_server);
 
     if ($opt->{'renew-check'}) {
         $opt->{'error'}->("Unsupported protocol for the renew check URL: $1.", 'RENEW_CHECK_URL') if ($opt->{'renew-check'}=~s~^(.*?)://~~ and uc($1) ne 'HTTPS');
@@ -562,8 +596,7 @@ sub process_challenge {
         }
         my $file = "$path/$challenge->{token}";
         if (-e $file) {
-           $challenge->{'logger'}->error("File already exists - this should not be the case, please move it or remove it.");
-           return 0;
+           $challenge->{'logger'}->warn("File already exists - might happen if previous validations failed and -unlink option was not used.");
         }
         if (_write($file, $text)) {
            $challenge->{'logger'}->error("Failed to save a challenge file '$file' for domain '$challenge->{domain}'");
@@ -637,7 +670,7 @@ sub process_verification_dns {
 
 sub usage_and_exit {
     my $opt = shift;
-    print "\n ZeroSSL Crypt::LE client v$VERSION\n\n";
+    print "\n Crypt::LE client v$VERSION\n\n";
     if ($opt->{'help'}) {
         print << 'EOF';
  ===============
@@ -707,13 +740,13 @@ e) To use basic DNS verification:
 f) To issue a wildcard certificate, which requires DNS verification:
 
  le.pl --key account.key --csr domain.csr --csr-key domain.key --crt domain.crt
-       --domains "*.domain.ext" --generate-missing --handle-as dns --api 2
+       --domains "*.domain.ext" --generate-missing --handle-as dns
 
 To include a "bare domain", add it too, since it is NOT covered by the wildcard:
 
  le.pl --key account.key --csr domain.csr --csr-key domain.key --crt domain.crt
         --domains "*.domain.ext,domain.ext" --generate-missing
-        --handle-as dns --api 2
+        --handle-as dns
 
 g) To just generate the keys and CSR:
 
@@ -815,11 +848,13 @@ EOF
 -complete-params <json|file> : JSON for the completion module (optional).
 -issue-code XXX              : Exit code to use on issuance/renewal (optional).
 -email <some@mail.address>   : Email for expiration notifications (optional).
--server <url|host>           : Use custom server URL (optional).
+-server <url|host>           : Custom server URL for API root (optional).
+-directory <url>             : Custom server URL for API directory (optional).
 -api <version>               : API version to use (optional).
 -update-contacts <emails>    : Update contact details.
 -export-pfx <password>       : Export PFX (Windows binaries only).
 -tag-pfx <tag>               : Tag PFX with a specific name.
+-alternative <num>           : Save an alternative ceritifcate (if available).
 -config <file>               : Configuration file for the client.
 -log-config <file>           : Configuration file for logging.
 -generate-missing            : Generate missing files (key, csr and csr-key).
